#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : NoMask
# @Time         : 2020/4/10 11:33 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import tensorflow as tf


class NoMask(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super(NoMask, self).__init__(**kwargs)

    def build(self, input_shape):
        # Be sure to call this somewhere!
        super(NoMask, self).build(input_shape)

    def call(self, x, mask=None, **kwargs):
        return x

    def compute_mask(self, inputs, mask=None):
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : gpu_lgb
# @Time         : 2019-07-04 12:14
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

"""默认参数https://www.kaggle.com/c/home-credit-default-risk/discussion/59347#latest-542068"""
import lightgbm as lgb

from sklearn.datasets import make_classification


X, y = make_classification(10000)
param = {
    'num_leaves': 10,
    'max_bin': 127,
    'min_data_in_leaf': 11,
    'learning_rate': 0.02,
    'min_sum_hessian_in_leaf': 0.00245,
    'bagging_fraction': 1.0,
    'bagging_freq': 5,
    'feature_fraction': 0.05,
    'lambda_l1': 4.972,
    'lambda_l2': 2.276,
    'min_gain_to_split': 0.65,
    'max_depth': 14,
    'save_binary': True,
    'seed': 1337,
    'feature_fraction_seed': 1337,
    'bagging_seed': 1337,
    'drop_seed': 1337,
    'data_random_seed': 1337,
    'objective': 'binary',
    'boosting_type': 'gbdt',
    'verbose': 1,
    'metric': 'auc',
    'is_unbalance': True,
    'boost_from_average': False,
    'device': 'gpu',
    'gpu_platform_id': 0,
    'gpu_device_id': 0,
    'gpu_use_dp':True
}

data = lgb.Dataset(X, y)
_ = lgb.cv(param, data, num_boost_round=10000, verbose_eval=1000)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : show_column
# @Time         : 2020/4/13 4:54 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import tensorflow as tf
from tensorflow.python.feature_column.feature_column import _LazyBuilder

def show_column(data: dict, feature_column):
    try:
        return show_column1(data, feature_column)
    except Exception as e:
        return show_column2(data, feature_column)


def show_column1(data: dict, feature_column):
    builder = _LazyBuilder(data)
    id_tensor, weight = feature_column._get_sparse_tensors(builder)

    if weight is None:
        return id_tensor.values
    else:
        return id_tensor.values, weight.values


def show_column2(data: dict, feature_column, feature_type='Seq'):
    """wrap a categorical column with an embedding_column or indicator_column

    :param feature_column: feature_column.numeric_column('age')
    :param batch_data: {'feature_name': [1,2,3,4]}
    :return:
    """
    if feature_type == 'Seq':
        feature_layer = tf.keras.experimental.SequenceFeatures([feature_column])
        sequence_input, sequence_length = feature_layer(data)
        return sequence_input, sequence_length
    else:
        feature_layer = tf.keras.layers.DenseFeatures([feature_column])
        return feature_layer(data).numpy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : DNN
# @Time         : 2020-03-13 13:42
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import tensorflow as tf


class DNN(tf.keras.layers.Layer):

    def __init__(self,
                 hidden_units_list: List[int] = (64, 32, 16, 4),
                 activation='relu',
                 kernel_regularizer=tf.keras.regularizers.l2(),
                 use_bn=False,
                 dropout_rate=0,
                 seed=666,
                 name='DNN',
                 **kwargs):
        super().__init__(name=name, **kwargs)

        self.hidden_units_list = hidden_units_list
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.kernel_regularizer = kernel_regularizer
        self.use_bn = use_bn
        self.seed = seed
        self.num_layer = len(self.hidden_units_list)

    def build(self, input_shape):
        super().build(input_shape)  # self.built = True

        self.dense_layers = []
        for index, units in enumerate(self.hidden_units_list):
            _ = tf.keras.layers.Dense(
                units,
                activation=self.activation,
                use_bias=True,
                kernel_initializer='glorot_uniform',
                bias_initializer='zeros',
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=None,
                name=f"dense{index}"
            )
            self.dense_layers.append(_)

        # BN
        if self.use_bn:
            self.bn_layers = [tf.keras.layers.BatchNormalization() for _ in range(self.num_layer)]

        self.dropout_layers = []
        for i in range(self.num_layer):
            _ = tf.keras.layers.Dropout(self.dropout_rate, seed=self.seed + i)
            self.dropout_layers.append(_)

    def call(self, inputs, **kwargs):
        """http://www.luyixian.cn/news_show_256709.aspx
        BN和Dropout共同使用时会出现的问题
        BN和Dropout单独使用都能减少过拟合并加速训练速度，但如果一起使用的话并不会产生1+1>2的效果，相反可能会得到比单独使用更差的效果。
        相关的研究参考论文：Understanding the Disharmony between Dropout and Batch Normalization by Variance Shift
        本论文作者发现理解 Dropout 与 BN 之间冲突的关键是网络状态切换过程中存在神经方差的（neural variance）不一致行为。
        试想若有图一中的神经响应 X，当网络从训练转为测试时，Dropout 可以通过其随机失活保留率（即 p）来缩放响应，并在学习中改变神经元的方差，
        而 BN 仍然维持 X 的统计滑动方差。这种方差不匹配可能导致数值不稳定（见下图中的红色曲线）。而随着网络越来越深，最终预测的数值偏差可能会累计，
        从而降低系统的性能。简单起见，作者们将这一现象命名为「方差偏移」。
        事实上，如果没有 Dropout，那么实际前馈中的神经元方差将与 BN 所累计的滑动方差非常接近（见下图中的蓝色曲线），这也保证了其较高的测试准确率。
        """
        deep_input = inputs
        for i in range(self.num_layer):
            fc = self.dense_layers[i](deep_input)  # 注意下次循环的输入
            fc = self.bn_layers[i](fc) if self.use_bn else fc
            fc = self.dropout_layers[i](fc)
            deep_input = fc

        return fc

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.hidden_units_list[-1],)

    def get_config(self):
        base_config = super().get_config()
        config = {
            'hidden_units': self.hidden_units_list,
            'activation': self.activation,
            'kernel_regularizer': self.kernel_regularizer,
            'use_bn': self.use_bn,
            'dropout_rate': self.dropout_rate,
            'seed': self.seed
        }
        return {**base_config, **config}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-09-10 15:21
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from .Dice import Dice

import tensorflow as tf
from tensorflow.keras.layers import Activation
from tensorflow.keras.utils import get_custom_objects


class Mish(Activation):
    '''
    Mish Activation Function.
    .. math::
        mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + e^{x}))
    Shape:
        - Input: Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.
        - Output: Same shape as the input.
    Examples:
        >>> X = Activation('Mish', name="conv1_act")(X_input)
    '''

    def __init__(self, activation, **kwargs):
        super(Mish, self).__init__(activation, **kwargs)
        self.__name__ = 'Mish'


def mish(inputs):
    return inputs * tf.math.tanh(tf.math.softplus(inputs))


get_custom_objects().update({'Mish': Mish(mish)})


def custom_gelu(x):
    return 0.5 * x * (1 + tf.tanh(tf.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3))))


get_custom_objects().update({'custom_gelu': Activation(custom_gelu)})

# 注册函数
funcs = [mish, custom_gelu]
custom_objects = {func.__name__: Activation(func) for func in funcs}

get_custom_objects().update(custom_objects)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : io_utils
# @Time         : 2020/4/21 12:04 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import os

get_module_path = lambda path, file=__file__: \
    os.path.normpath(os.path.join(os.getcwd(), os.path.dirname(file), path))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : Linear
# @Time         : 2020/4/10 11:34 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf


class Linear(tf.keras.layers.Layer):

    def __init__(self, l2_reg=0.0, mode=0, use_bias=False, name='Linear', **kwargs):
        super().__init__(name=name, **kwargs)

        assert mode in (0, 1, 2), ValueError("mode must be 0,1 or 2")  # raise ValueError("mode must be 0,1 or 2")

        self.l2_reg = l2_reg
        self.mode = mode
        self.use_bias = use_bias

        if self.use_bias:
            self.bias = self.add_weight(name='linear_bias',
                                        shape=(1,),
                                        initializer=tf.keras.initializers.Zeros(),
                                        trainable=True)

    def build(self, input_shape):
        super().build(input_shape)  # self.built = True

        if self.mode in (1, 2):
            shape = [int(input_shape[-1]), 1] if self.mode == 1 else [int(input_shape[1][-1]), 1]
            self.kernel = self.add_weight(
                'linear_kernel',
                shape=shape,
                initializer=tf.keras.initializers.glorot_normal(),
                regularizer=tf.keras.regularizers.l2(self.l2_reg),
                trainable=True)

    def call(self, inputs, **kwargs):
        if self.mode == 0:
            sparse_input = inputs
            linear_logit = tf.reduce_sum(sparse_input, axis=-1, keep_dims=True)
        elif self.mode == 1:
            dense_input = inputs
            fc = tf.tensordot(dense_input, self.kernel, axes=(-1, 0))
            linear_logit = fc
        else:
            sparse_input, dense_input = inputs
            fc = tf.tensordot(dense_input, self.kernel, axes=(-1, 0))
            linear_logit = tf.reduce_sum(sparse_input, axis=-1, keep_dims=False) + fc
        if self.use_bias:
            linear_logit += self.bias

        return linear_logit

    def compute_output_shape(self, input_shape):
        return (None, 1)

    def compute_mask(self, inputs, mask=None):
        return None

    def get_config(self, ):
        base_config = super(Linear, self).get_config()
        config = {'mode': self.mode, 'l2_reg': self.l2_reg, 'use_bias': self.use_bias}
        return {**base_config, **config}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : __init__.py
# @Time         : 2020/5/22 1:13 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


from collections import namedtuple


class Column(object):
    pass


class NumericColumn(Column,
                    namedtuple('NumericColumn', ('name', 'shape', 'dtype', 'normalizer_fn'))):
    def __new__(cls, name, shape=(1,), normalizer_fn=None):
        _ = super().__new__(cls, name=name,
                            shape=shape,
                            dtype='float32',
                            normalizer_fn=normalizer_fn)
        return _


class CategoricalColumn(Column,
                        namedtuple('CategoricalColumn',
                                   ('name', 'shape', 'vocabulary_size', 'embedding_dim', 'embedding_name',
                                    'field', 'dtype'))):
    def __new__(cls, name='CategoricalColumn',
                vocabulary_size=1000,  # nunique
                embedding_dim=64,
                embedding_name=None,
                field='field'):

        if embedding_name is None:
            embedding_name = name

        if embedding_dim == 'auto' or embedding_dim is None:
            embedding_dim = int(6 * vocabulary_size ** 0.25)

        _ = super().__new__(cls, name=name,
                            vocabulary_size=vocabulary_size,
                            embedding_dim=embedding_dim,
                            embedding_name=embedding_name,
                            field=field,
                            dtype='int32',
                            shape=(1,))
        return _


class SequenceCategoricalColumn(Column,
                                namedtuple('SequenceCategoricalColumn',
                                           ('name', 'shape', 'maxlen', 'combiner',
                                            'vocabulary_size', 'embedding_dim', 'embedding_name',
                                            'length_name', 'weight_name',
                                            'weight_norm', 'field', 'dtype'))):
    def __new__(cls, name,
                maxlen,
                vocabulary_size,
                embedding_dim=None,
                embedding_name=None,
                combiner='mean',
                length_name="length_name",
                weight_name="weight_name",
                weight_norm=None,
                field='field'):

        if embedding_name is None:
            embedding_name = name

        if embedding_dim == 'auto' or embedding_dim is None:
            embedding_dim = int(6 * vocabulary_size ** 0.25)

        _ = super().__new__(cls, name=name,
                            maxlen=maxlen,
                            vocabulary_size=vocabulary_size,
                            embedding_dim=embedding_dim,
                            embedding_name=embedding_name,
                            combiner=combiner,
                            length_name=length_name,
                            weight_name=weight_name,
                            weight_norm=weight_norm,
                            field=field,
                            dtype='int32',
                            shape=(maxlen,))
        return _



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-09-10 14:48
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

from .pipe import *
from .logger import *
from .timer import timer
from .cprint import cprint
from .NT import NT

if os.environ.get('disable_eager_execution'):
    import tensorflow as tf

    tf.compat.v1.disable_eager_execution()
    logger.info('关闭动态图')


# from distutils.util import strtobool
# import numpy as np
# import tensorflow as tf
#
# # 判断是tf.keras还是纯keras的标记
# is_tf_keras = strtobool(os.environ.get('TF_KERAS', '0'))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-09-10 15:12
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


def focal_loss_fixed(y_true, y_pred):
    pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
    pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))
    return -K.mean(alpha * K.pow(1. - pt_1, gamma) * K.log(K.epsilon()+pt_1))-K.mean((1-alpha) * K.pow( pt_0, gamma) * K.log(1. - pt_0 + K.epsilon()))

get_custom_objects().update({'focal_loss_fn': focal_loss()})

get_custom_objects
import tensorflow as tf
tf.keras.losses.Loss
tf.keras.utils.get_custom_objects()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : fc
# @Time         : 2020/5/21 4:42 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from . import *

from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import tensorflow as tf
from tensorflow.python.feature_column.feature_column_v2 import \
    FeatureColumn, NumericColumn, CategoricalColumn, SequenceCategoricalColumn, EmbeddingColumn, \
    BucketizedColumn


def dense_feature_out(fc: FeatureColumn, data: dict):
    return tf.keras.layers.DenseFeatures([fc])(data)


def fc2input(fc: FeatureColumn):
    return tf.keras.layers.Input(shape=fc.shape, name=fc.name, dtype=fc.dtype)


#################################

def get_linear_output(fc2input, dense_dim=1, dense_activation=None):
    dense_out = list(get_dense_out(fc2input, dense_dim=dense_dim, dense_activation=dense_activation))
    embedding_out = list(get_embedding_out(fc2input, 1))
    return tf.keras.layers.concatenate(dense_out + embedding_out)


def get_dense_out(fc2input, dense_dim=1, dense_activation=None):
    for fc, input in fc2input:
        if isinstance(fc, NumericColumn):
            _ = tf.keras.layers.Dense(dense_dim, activation=dense_activation)(input)
            yield _


def get_embedding_out(fc2input, embedding_dim=None):
    for fc, input in fc2input:
        if isinstance(fc, CategoricalColumn):
            _ = tf.keras.layers.Embedding(fc.vocabulary_size,
                                          fc.embedding_dim if embedding_dim is None else embedding_dim,
                                          mask_zero=True)(input)
            if embedding_dim == 1:
                _ = tf.keras.layers.Flatten()(_)
            yield _

        elif isinstance(fc, SequenceCategoricalColumn):
            _ = tf.keras.layers.Embedding(fc.vocabulary_size,
                                          fc.embedding_dim if embedding_dim is None else embedding_dim,
                                          mask_zero=True)(input)

            if embedding_dim == 1:
                if fc.combiner == 'mean':
                    _ = tf.keras.layers.GlobalAveragePooling1D()(_)
                    yield _

                elif fc.combiner == 'max':
                    _ = tf.keras.layers.GlobalMaxPooling1D()(_)
                    yield _
                else:
                    raise ValueError(f'Unsupported combiner {fc.combiner}')
            else:
                yield _



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : _layers
# @Time         : 2020/4/10 11:58 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

import tensorflow as tf


class Linear(tf.keras.layers.Layer):
    def __init__(self, units=32, **kwargs):
        super(Linear, self).__init__(**kwargs)
        self.units = units

    # build方法一般定义Layer需要被训练的参数。
    def build(self, input_shape):
        self.w = self.add_weight(shape=(input_shape[-1], self.units),
                                 initializer='random_normal',
                                 trainable=True)
        self.b = self.add_weight(shape=(self.units,),
                                 initializer='random_normal',
                                 trainable=True)
        # self.built = True
        super(Linear, self).build(input_shape)  # 相当于设置self.built = True

    # call方法一般定义正向传播运算逻辑，__call__方法调用了它。
    def call(self, inputs, **kwargs):
        return tf.matmul(inputs, self.w) + self.b

    # 如果要让自定义的Layer通过Functional API 组合成模型时可以序列化，需要自定义get_config方法。
    def get_config(self):
        config = super(Linear, self).get_config()
        config.update({'units': self.units})
        return config

    @classmethod
    def test_compute_output_shape(cls, input_shape=(None, 16)):
        """若推断失败，需重写`compute_output_shape`"""
        layer = cls()
        layer.build(input_shape)
        _ = layer.compute_output_shape(input_shape)
        print(_)

        # 如果built = False，调用__call__时会先调用build方法, 再调用call方法。
        # layer(tf.random.uniform((66, input_shape[-1])))


if __name__ == '__main__':
    print(Linear.test_compute_output_shape((100,)))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : EmbeddingPooling
# @Time         : 2020/5/22 4:12 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf

K = tf.keras.backend


class EmbeddingPooling(tf.keras.layers.Layer):
    def __init__(self, input_dim,
                 output_dim,
                 combiner=None,
                 mask_zero=True,
                 input_length=None,
                 name='EmbeddingPooling',
                 **kwargs):
        super().__init__(name=name, **kwargs)

        self.input_dim = input_dim
        self.output_dim = output_dim  # output_dim==1 要压平
        self.combiner = combiner
        self.mask_zero = mask_zero
        self.input_length = input_length

    def build(self, input_shape):
        super().build(input_shape)  # self.built = True

        # embeddings_initializer = 'uniform',
        # embeddings_regularizer = None,
        # activity_regularizer = None,

        self.embedding_layer = tf.keras.layers.Embedding(
            input_dim=self.input_dim,
            output_dim=self.output_dim,
            mask_zero=self.mask_zero,
            input_length=self.input_length)

        if self.combiner:
            if self.combiner == 'max':
                self.pool_layer = tf.keras.layers.GlobalMaxPooling1D()

            elif self.combiner == 'mean':
                self.pool_layer = tf.keras.layers.GlobalAveragePooling1D()

            else:
                raise ValueError(f"Unsupported combiner {self.combiner}")

    def call(self, inputs, **kwargs):
        _ = self.embedding_layer(inputs)
        if self.combiner:
            return self.pool_layer(_)
        else:
            return _

    def compute_output_shape(self, input_shape):
        if self.combiner:
            return (None, self.output_dim)
        else:
            return (None, self.input_dim, self.output_dim)

    def get_config(self, ):
        base_config = self.get_config()
        config = {
            "input_dim": self.input_dim,
            "output_dim": self.output_dim,
            "combiner": self.combiner,
            "mask_zero": self.mask_zero,
            "input_length": self.input_length
        }
        return {**base_config, **config}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'OptimizedRounder'
__author__ = 'JieYuan'
__mtime__ = '2019/4/4'
"""
import numpy as np
import pandas as pd
import scipy as sp

from functools import partial
from sklearn.metrics import cohen_kappa_score


# https://www.kaggle.com/naveenasaithambi/optimizedrounder-improved
class OptimizedRounder(object):
    def __init__(self):
        self.coef_ = 0

    def _kappa_loss(self, coef, X, y):
        preds = pd.cut(X, [-np.inf] + list(np.sort(coef)) + [np.inf], labels=[0, 1, 2, 3, 4])
        return -cohen_kappa_score(y, preds, weights='quadratic')

    def fit(self, X, y):
        loss_partial = partial(self._kappa_loss, X=X, y=y)
        initial_coef = np.percentile(X, [2.73, 23.3, 50.3, 72])  # <= improved
        self.coef_ = sp.optimize.minimize(loss_partial, initial_coef, method='nelder-mead')

    def predict(self, X, coef):
        preds = pd.cut(X, [-np.inf] + list(np.sort(coef)) + [np.inf], labels=[0, 1, 2, 3, 4])
        return preds

    def coefficients(self):
        return self.coef_['x']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : fc
# @Time         : 2020-03-11 16:46
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import tensorflow as tf
from tensorflow.python.keras.layers import \
    Dense, Conv1D, GlobalMaxPooling1D, Concatenate, Dropout, concatenate



# #!/usr/bin/env python
# # -*- coding: utf-8 -*-
# # @Project      : inn.
# # @File         : Emebedding
# # @Time         : 2020/5/18 4:12 下午
# # @Author       : yuanjie
# # @Email        : yuanjie@xiaomi.com
# # @Software     : PyCharm
# # @Description  :
#
#
import tensorflow as tf

tf.sparse.to_dense

#
# from tensorflow.keras.layers import *
#
#
# def get_embedding(vocabulary_size, embedding_dim, maxlen):
#     """
#       Input shape:
#     2D tensor with shape: `(batch_size, input_length)`.
#
#   Output shape:
#     3D tensor with shape: `(batch_size, input_length, output_dim)`.
#     """
#
#     Embedding(input_dim=vocabulary_size,
#               output_dim=embedding_dim,
#               embeddings_initializer,
#               embeddings_regularizer,
#               name='embedding',
#               mask_zero
#               input_length=maxlen)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : hashtrick
# @Time         : 2020-04-10 16:54
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import hashlib as _hashlib
from sklearn.utils.murmurhash import murmurhash3_32 as _murmurhash3_32

"""Java
import java.nio.charset.StandardCharsets
import com.google.common.hash.Hashing.murmur3_32

def hash(key: String = "key", value: String = "value", bins: Int = 10000): Int = {
    val hashValue: Int = murmur3_32.newHasher.putString(f"{key}:{value}", StandardCharsets.UTF_8).hash.asInt
    Math.abs(hashValue) % bins  
  }
"""


def _md5(string: str):
    return _hashlib.md5(string.encode('utf8')).hexdigest()


def murmurhash(key="key", value="value", bins=None, str2md5=False):
    """key:value"""
    string = f"{key}:{value}"
    if str2md5:
        string = _md5(string)

    _ = abs(_murmurhash3_32(string))
    return _ % bins if bins else _


if __name__ == '__main__':
    print(_md5("key:value"))
    print(murmurhash())
    print(murmurhash(str2md5=True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : xgb_convert
# @Time         : 2020-03-16 13:15
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
# https://github.com/amesar/mlflow-examples/blob/220e03d6457ac9a2ac08589a96881691679678a3/scala/onnx/README.md
# https://github.com/amesar/mlflow-examples/tree/220e03d6457ac9a2ac08589a96881691679678a3
# https://github.com/onnx/tutorials/blob/master/tutorials/PytorchTensorflowMnist.ipynb


from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split

iris = load_iris()
X, y = iris.data, iris.target
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.8, random_state=42)

import xgboost as xgb

xgb_model = xgb.XGBClassifier()
xgb_model = xgb_model.fit(X_train, y_train)

print("Test data accuracy of the xgb classifier is {:.2f}".format(xgb_model.score(X_test, y_test)))

from onnxmltools.convert import convert_xgboost, convert_lightgbm
from onnxconverter_common.data_types import FloatTensorType

onnx_model = convert_xgboost(xgb_model, initial_types=[("input", FloatTensorType([1, 4]))])

with open("gbtree.onnx", "wb") as f:
    f.write(onnx_model.SerializeToString())

if __name__ == '__main__':
    from inference import Inference

    infer = Inference("gbtree.onnx")
    print(infer.run(X[:1]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : get_namedtuple
# @Time         : 2020/4/13 12:04 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from . import namedtupled


class NT(object):

    def __new__(cls, _nt_name="NT", **kwargs):
        nt = namedtupled.map(kwargs, _nt_name=_nt_name)
        return nt


if __name__ == '__main__':
    class SF(NT):

        def __new__(cls, a='a', b='b'):

            return super().__new__(cls, a=a, b=b)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : Dice
# @Time         : 2020/5/13 8:55 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import tensorflow as tf
from tensorflow.python.keras.initializers import Zeros


class Dice(tf.keras.layers.Layer):
    """The Data Adaptive Activation Function in DIN,which can be viewed as a generalization of PReLu and can adaptively adjust the rectified point according to distribution of input data.

      Input shape
        - Arbitrary. Use the keyword argument `input_shape` (tuple of integers, does not include the samples axis) when using this layer as the first layer in a model.

      Output shape
        - Same shape as the input.

      Arguments
        - **axis** : Integer, the axis that should be used to compute data distribution (typically the features axis).

        - **epsilon** : Small float added to variance to avoid dividing by zero.

      References
        - [Zhou G, Zhu X, Song C, et al. Deep interest network for click-through rate prediction[C]//Proceedings of the 24th ACM SIGKDD International Conference on Knowledge Discovery & Data Mining. ACM, 2018: 1059-1068.](https://arxiv.org/pdf/1706.06978.pdf)
    """

    def __init__(self, axis=-1, epsilon=1e-9, **kwargs):
        self.axis = axis
        self.epsilon = epsilon
        super(Dice, self).__init__(**kwargs)

    def build(self, input_shape):
        super(Dice, self).build(input_shape)  # Be sure to call this somewhere!

        self.uses_learning_phase = True  # 学习阶段

        self.bn = tf.keras.layers.BatchNormalization(
            xis=self.axis,
            epsilon=self.epsilon,
            center=False,
            scale=False)  # TODO:为啥是false

        self.alphas = self.add_weight(
            shape=(input_shape[-1],),
            initializer=Zeros(),
            dtype=tf.float32,
            name='dice_alpha')  # name='alpha_'+self.name

    def call(self, inputs, training=None, **kwargs):
        """
        有些Keras层,如BN,Dropout,在训练和测试过程中的行为不一致, 你可以通过打印layer.uses_learning_phase来确定当前层工作在训练模式还是测试模式
        if training is None:
            tf.keras.backend.learning_phase() # TODO:打印一下
        """
        inputs_normed = self.bn(inputs, training=training)
        x_p = tf.sigmoid(inputs_normed)
        return x_p * inputs + (1.0 - x_p) * self.alphas * inputs

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        base_config = super().get_config()
        config = {'axis': self.axis, 'epsilon': self.epsilon}
        return {**base_config, **config}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'rank_encoder'
__author__ = 'JieYuan'
__mtime__ = '19-3-11'
"""
from collections import OrderedDict

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class CountRankEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, topn=None):
        """

        :param topn: 仅保留topn个类别
        """
        self.topn = topn
        self.mapper = None

    def fit(self, y):
        ce = pd.Series(y).value_counts(True, dropna=False)  # 计数编码
        if self.topn:
            ce = ce[:self.topn]
            print(f"Coverage: {ce.sum() * 100:.2f}%")

        self.mapper = ce.rank(method='first').to_dict(OrderedDict)  # rank 合理？
        return self

    def transform(self, y):
        """不在训练集的补0，不经常出现补0"""
        return pd.Series(y).map(self.mapper).fillna(0)


if __name__ == '__main__':
    import numpy as np

    s = ['a', 'a', 'b', 'b', 'c'] + [np.nan] * 6
    re = CountRankEncoder(2)

    print(re.fit_transform(s))
    print(re.mapper)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : muit_inputs_outputs
# @Time         : 2020/5/18 9:31 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :


import tensorflow as tf
from tensorflow.keras.layers import *
from sklearn.datasets import load_iris

from inn.layers import DNN
from inn.dateset import Dataset

X, y = load_iris(True)
# 构建数据集
ds = Dataset().from_cache((X, X), (y, y))
ds_test = Dataset().from_cache(inputs=(X, X))

for i in ds:
    print(i)
    break

for i in ds_test:
    print(i)
    break

# 搭建模型
input1 = Input((4,))
input2 = Input((4,))
x = tf.keras.layers.concatenate([input1, input2])
x = DNN([64, 16, 4])(x)
output1 = Dense(1, 'sigmoid', name='fc1')(x)
output2 = Dense(1, 'sigmoid', name='fc2')(x)

model = tf.keras.Model(inputs=[input1, input2], outputs=[output1, output2])

# 目标函数定义，需与输出层名字对应
losses = {'fc1': 'categorical_crossentropy',
          'fc2': 'categorical_crossentropy'}

loss_weights = {'fc1': 0.9,
                'fc2': 0.1}

model.compile(loss=losses, loss_weights=loss_weights, metrics=['accuracy'])

model.fit(ds, epochs=1)  # model.fit(ds, epochs=1, steps_per_epoch=100)
print(model.predict(ds_test))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : RankEncoder
# @Time         : 2019-09-20 18:51
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 



import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class RankEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, method='average', na_option='keep'):
        """
        method : {'average', 'min', 'max', 'first', 'dense'}, default 'average'
            How to rank the group of records that have the same value
            (i.e. ties):

            * average: average rank of the group
            * min: lowest rank in the group
            * max: highest rank in the group
            * first: ranks assigned in order they appear in the array
            * dense: like 'min', but rank always increases by 1 between groups
        numeric_only : bool, optional
            For DataFrame objects, rank only numeric columns if set to True.
        na_option : {'keep', 'top', 'bottom'}, default 'keep'
            How to rank NaN values:

            * keep: assign NaN rank to NaN values
            * top: assign smallest rank to NaN values if ascending
            * bottom: assign highest rank to NaN values if ascending
        ascending : bool, default True
            Whether or not the elements should be ranked in ascending order.
        """
        self.method = method
        self.na_option = na_option

    def transform(self, y):
        """不在训练集的补0，不经常出现补0"""
        return pd.Series(y).rank(method=self.method, na_option=self.na_option)  # .fillna(0)


if __name__ == '__main__':
    import numpy as np

    s = ['a', 'a', 'b', 'b', 'c'] + [np.nan] * 6
    re = RankEncoder()

    print(re.transform(s))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : data_processor_utils
# @Time         : 2020/4/21 12:04 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import numpy as np

_norm = lambda x, ord=1: x / np.linalg.norm(x, ord, axis=len(x.shape) > 1, keepdims=True)


def l1(x):
    return _norm(x)


def l2(x):
    return _norm(x, 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : torch_convert
# @Time         : 2020-03-16 15:57
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import numpy as np
import torch
from torch import nn, onnx
import torch.nn.functional as F
from sklearn.datasets import load_iris

iris = load_iris()
X, y = iris.data, iris.target


class NN(nn.Module):
    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(4, 16)
        self.fc2 = nn.Linear(16, 3)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = self.fc2(x)  # F.softmax
        return x


net = NN()
print(net)

# Export: X.shape
onnx.export(
    net, torch.randn(128, 4), './iris.onnx', verbose=True,
    input_names=['input_name'],
    output_names=['output_name']
)

if __name__ == '__main__':
    from inference import Inference

    infer = Inference()
    print(infer.run(X[:128])[:5])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : __init__.py
# @Time         : 2020-03-12 12:48
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : https://github.com/wandb/client

import tensorflow as tf
import keras
import keras.backend as K
from keras.models import *
from keras.layers import *
from keras.layers import Input, Embedding, Dense, Dropout, BatchNormalization, Flatten, Concatenate, \
    Reshape, Activation
from keras.optimizers import SGD
from keras.callbacks import ModelCheckpoint, LearningRateScheduler, Callback

from keras.utils import np_utils

# from tensorflow.keras.datasets import fashion_mnist
# (X_train_orig, y_train_orig), (X_test, y_test) = fashion_mnist.load_data() # (60000, 28, 28)


from sklearn.datasets import make_regression, make_classification

X, y = make_classification(10240, 16)

# 789ae399af943555652e476ff1d0c0452ee86564
import wandb
from wandb.keras import WandbCallback  # torch

wandb.init(project="Keraser")
config = wandb.config

# Track hyperparameters
config.dropout = 0.2
config.hidden_layer_size = 128
config.layer_1_size = 16
config.layer_2_size = 32
config.learn_rate = 0.01
config.decay = 1e-6
config.momentum = 0.9
config.epochs = 8

input_shape = X.shape[1:]
input = Input(input_shape)
x = input
for i in [128, 64, 32, 16, 1]:
    x = Dense(i, activation='relu')(x)

output_model = Activation('sigmoid')(x)

model = Model(inputs=input, outputs=output_model)
model.summary()

model.compile('sgd', loss='binary_crossentropy', metrics=['accuracy', keras.metrics.AUC()])
model.fit(X, y, 128, 10)


tf.keras.losses.mean_squared_error,
import tensorflow_addons as tfa

tfa


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : DataLoader
# @Time         : 2019-10-21 13:01
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
"""
cache
"""
import numpy as np
import pandas as pd
import tensorflow as tf

from functools import partial


class Dataset(object):
    """
    https://tensorflow.google.cn/guide/data?hl=zh_cn

    """

    def __init__(self, batchsize=128, cache_filename=""):

        self.batchsize = batchsize
        self.cache_filename = cache_filename

    def from_cache(self, inputs, outputs=None, is_test=False, shuffle_buffer_size=10000, shuffle_seed=None):
        """
        多输入多输出inputs/outputs对应元组
        """
        if isinstance(inputs, tuple):  # Multi-inputs
            pass
        elif isinstance(inputs, (list, np.ndarray, dict, pd.DataFrame)):
            if isinstance(inputs, pd.DataFrame):
                inputs = inputs.to_dict('list')
        else:
            raise ValueError("`inputs` Data Type Error")

        if outputs is None:
            tensors = (inputs,)
        else:
            tensors = (inputs, outputs)

        # Common
        ds = tf.data.Dataset.from_tensor_slices(tensors)

        if outputs is None or is_test:  # 避免测试集shuffle
            pass
        else:
            ds = ds.shuffle(shuffle_buffer_size, seed=shuffle_seed)

        ds = ds.batch(self.batchsize)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)  # todo: .repeat() 更乱一些？？？在每个epoch内将图片打乱组成大小为32的batch，并重复10次。

        return ds

    def from_generator(self):
        # TODO: 增加对文件的操作（txt/tfrecord）
        # tf.data.Dataset.from_generator()
        pass

    def _from_np_array(self, array):
        # Common
        buffer_size = len(array)
        ds = tf.data.Dataset.from_tensor_slices(array)
        ds = ds.shuffle(buffer_size, seed=None)
        ds = ds.batch(self.batchsize)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)

        return ds

    def _from_pd_dataframe(self, df: pd.DataFrame, label="label"):
        """
        import tensorflow as tf
        dataset = tf.data.Dataset.from_tensor_slices(({"a": [1, 2], "b": [3, 4]}, [0, 1]))
        print(list(dataset.as_numpy_iterator()))

        :param df:
        :param label:
        :return:
        """
        if label and label in df.columns:
            df = df.drop(labels=[label], axis=1)
            labels = df[label]
            tensors = (df.to_dict('list'), labels)  # df.to_dict('series')
        else:
            tensors = df.to_dict('list')

        # Common
        buffer_size = len(df)
        ds = tf.data.Dataset.from_tensor_slices(tensors)
        ds = ds.shuffle(buffer_size, seed=None)
        ds = ds.batch(self.batchsize)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)

        # features_dataset = tf.data.Dataset.from_tensor_slices(features)
        # labels_dataset = tf.data.Dataset.from_tensor_slices(labels)
        # tf.data.Dataset.zip((features_dataset, labels_dataset))
        return ds

    def from_tfrecord(self,
                      feature_dict: dict,
                      file_pattern: str = None,
                      label='label',
                      file_shuffle=True,
                      file_shuffle_seed=666,
                      shuffle_buffer_size=0):
        """注意缺失值问题
        ds = Dataset()
        feature_dict = {
            'id': tf.io.FixedLenFeature((), tf.int64, default_value=0),
            'feature': tf.io.FixedLenFeature((), tf.int64, default_value=0) # default_value=tf.zeros([], dtype=tf.float32)
            }
        ds = ds.from_tfrecord(feature_dict, '/Users/yuanjie/Desktop/Projects/Spark/MIPush/test-output.tfrecord/part*')

        :param feature_dict:
        :param file_pattern:
        :param shuffle:
        :param seed:
        :param shuffle_buffer_size:
        :return:
        """
        # TODO: cache
        # assert isinstance(filename, str), f"file path error: {filename}"
        #
        # if Path(filename).is_dir():
        #     filename = list(map(str, Path(filename).glob(glob_regex))) # tf.data.Dataset.list_files

        # parser_fn = partial(tf.io.parse_single_example, features=feature_dict)
        parser_fn = partial(self._tfrecord_parser_fn, features=feature_dict, label=label)

        filenames = tf.data.Dataset.list_files(file_pattern, file_shuffle, file_shuffle_seed)

        # ds
        ds = tf.data.TFRecordDataset(filenames, num_parallel_reads=tf.data.experimental.AUTOTUNE)
        ds = ds.map(parser_fn)

        if shuffle_buffer_size > 0:
            ds = ds.shuffle(shuffle_buffer_size, seed=None, reshuffle_each_iteration=True)

        ds = ds.batch(self.batchsize)
        ds = ds.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)  # 若内存泄露，需手动指定

        return ds

    def _tfrecord_parser_fn(self, example_photo, features, label='label'):
        _ = tf.io.parse_single_example(example_photo, features=features)

        return _, _.pop(label)  # X, y



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : mcdropout
# @Time         : 2020/5/13 7:55 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import numpy as np
import tensorflow as tf

(x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()
x_train, x_test = x_train / 255.0, x_test / 255.0
inp = tf.keras.layers.Input(shape=(28, 28))
x = tf.keras.layers.Flatten()(inp)
x = tf.keras.layers.Dense(512, activation=tf.nn.relu)(x)
x = tf.keras.layers.Dropout(0.5)(x, training=True)  # dropout 在训练和测试时都将开着
out = tf.keras.layers.Dense(10, activation=tf.nn.softmax)(x)
model = tf.keras.Model(inp, out)
model.compile(optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
model.fit(x_train, y_train, epochs=3)  # 在测试过程，dropout 也是打开的，得到的结果将会有波动，而不是完全一致

# dropout 层一直处于打开的状态，测试过程重复进行多次。
for _ in range(10):
    r = model.predict(x_test[:1])
    idx = np.argmax(r)
    print(idx, r[:, idx])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : pooling
# @Time         : 2020/5/20 6:23 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf

# 类似np
tf.keras.layers.GlobalMaxPooling1D
tf.keras.layers.GlobalAveragePooling1D

# 步长池化
tf.keras.layers.MaxPooling1D
tf.keras.layers.AveragePooling1D





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : FM
# @Time         : 2020/4/21 11:28 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


#############################################################################################################
# Created by qqgeogor
# https://www.kaggle.com/qqgeogor
#############################################################################################################

import numpy as np
from sklearn.base import BaseEstimator
from keras.layers import Input, Embedding, Dense, Flatten, merge, Activation
from keras.models import Model
from keras.regularizers import l2 as l2_reg
import itertools


def build_model(max_features, K=8, solver='adam', l2=0.0, l2_fm=0.0):
    inputs = []
    flatten_layers = []
    columns = range(len(max_features))
    for c in columns:
        inputs_c = Input(shape=(1,), dtype='int32', name='input_%s' % c)
        num_c = max_features[c]

        embed_c = Embedding(
            num_c,
            K,
            input_length=1,
            name='embed_%s' % c,
            W_regularizer=l2_reg(l2_fm)
        )(inputs_c)

        flatten_c = Flatten()(embed_c)

        inputs.append(inputs_c)
        flatten_layers.append(flatten_c)

    fm_layers = []
    for emb1, emb2 in itertools.combinations(flatten_layers, 2):
        dot_layer = merge([emb1, emb2], mode='dot', dot_axes=1)
        fm_layers.append(dot_layer)

    for c in columns:
        num_c = max_features[c]
        embed_c = Embedding(
            num_c,
            1,
            input_length=1,
            name='linear_%s' % c,
            W_regularizer=l2_reg(l2)
        )(inputs[c])

        flatten_c = Flatten()(embed_c)

        fm_layers.append(flatten_c)

    flatten = merge(fm_layers, mode='sum')
    outputs = Activation('sigmoid', name='outputs')(flatten)

    model = Model(input=inputs, output=outputs)

    model.compile(
        optimizer=solver,
        loss='binary_crossentropy'
    )

    return model


class KerasFM(BaseEstimator):
    def __init__(self, max_features=[], K=8, solver='adam', l2=0.0, l2_fm=0.0):
        self.model = build_model(max_features, K, solver, l2=l2, l2_fm=l2_fm)

    def fit(self, X, y, batch_size=128, nb_epoch=10, shuffle=True, verbose=1, validation_data=None):
        self.model.fit(X, y, batch_size=batch_size, nb_epoch=nb_epoch, shuffle=shuffle, verbose=verbose,
                       validation_data=None)

    def fit_generator(self, X, y, batch_size=128, nb_epoch=10, shuffle=True, verbose=1, validation_data=None,
                      callbacks=None):
        tr_gen = batch_generator(X, y, batch_size=batch_size, shuffle=shuffle)
        if validation_data:
            X_test, y_test = validation_data
            te_gen = batch_generator(X_test, y_test, batch_size=batch_size, shuffle=False)
            nb_val_samples = X_test[-1].shape[0]
        else:
            te_gen = None
            nb_val_samples = None

        self.model.fit_generator(
            tr_gen,
            samples_per_epoch=X[-1].shape[0],
            nb_epoch=nb_epoch,
            verbose=verbose,
            callbacks=callbacks,
            validation_data=te_gen,
            nb_val_samples=nb_val_samples,
            max_q_size=10
        )

    def predict(self, X, batch_size=128):
        y_preds = predict_batch(self.model, X, batch_size=batch_size)
        return y_preds


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : demo
# @Time         : 2019-09-10 15:17
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

# coding=utf-8
from tensorflow.keras.models import *
from tensorflow.keras.layers import *
import tensorflow.keras.backend as K

x_in = Input(shape=(784,))
x = x_in
x = Dense(100, activation='relu')(x)
x = Dense(784, activation='sigmoid')(x)

model = Model(x_in, x)
loss = K.mean((x - x_in) ** 2)
model.add_loss(loss)
model.compile(optimizer='adam') # from tensorflow_addons.optimizers import AdamW

model.summary()
# model.fit(x_train, None, epochs=5)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : keras_utils
# @Time         : 2020/4/21 8:22 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from tensorflow.keras import backend as K


def K_eval(x, backend=K):
    K = backend
    try:
        return K.get_value(K.to_dense(x))
    except Exception as e:
        try:
            eval_fn = K.function([], [x])
            return eval_fn([])[0]
        except Exception as e:
            return K.eager(K.eval)(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : FM
# @Time         : 2020/4/21 1:40 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf

K = tf.keras.backend


class FM(tf.keras.layers.Layer):
    """Factorization Machine models pairwise (order-2) feature interactions
     without linear term and bias.

      Input shape
        - 3D tensor with shape: ``(batch_size,field_size,embedding_size)``.

      Output shape
        - 2D tensor with shape: ``(batch_size, 1)``.

      References
        - [Factorization Machines](https://www.csie.ntu.edu.tw/~b97053/paper/Rendle2010FM.pdf)
    """

    def __init__(self, name='FM', **kwargs):

        super(FM, self).__init__(name=name, **kwargs)

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise ValueError(f"Unexpected inputs dimensions {len(input_shape)}, expect to be 3 dimensions")

        super(FM, self).build(input_shape)  # 相当于 self.built = True

    def call(self, inputs, **kwargs):
        """

        :param inputs:
        fm_input = np.array([
                    [[1, 3], [2, 3]],
                    [[1, 3], [2, 3]],
                    [[1, 3], [2, 3]]
                    ], dtype=np.float32) # (3, 2, 2)
        :param kwargs:
        :return:
        """

        if K.ndim(inputs) != 3:
            raise ValueError(f"Unexpected inputs dimensions {K.ndim(inputs)}, expect to be 3 dimensions")

        fm_input = inputs  # [None, SeqLen/field_size, EmbeddingDim]

        square_of_sum = tf.pow(tf.reduce_sum(fm_input, axis=1, keepdims=True), 2)  # [None, 1, EmbeddingDim]
        sum_of_square = tf.reduce_sum(fm_input * fm_input, axis=1, keepdims=True)  # [None, 1, EmbeddingDim]
        cross_term = square_of_sum - sum_of_square  # [None, 1, EmbeddingDim]
        cross_term = 0.5 * tf.reduce_sum(cross_term, axis=2, keepdims=False)  # [None, 1]

        return cross_term

    def compute_output_shape(self, input_shape):
        return (None, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = '__init__.py'
__author__ = 'JieYuan'
__mtime__ = '19-1-10'
"""

###################################

from .MultiHot import MultiHot

###################################
from .CountEncoder import CountEncoder
from .RankEncoder import RankEncoder
from .CountRankEncoder import CountRankEncoder
from .OutlierEncoder import OutlierEncoder
from .DataAugment import DataAugment



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : utils
# @Time         : 2020-02-19 11:46
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

import tensorflow as tf
import pandas as pd
from functools import partial


def libsvm2df(file, max_index=16, batch_size=128):
    dic = dict(zip(range(max_index), [0] * max_index))
    fn = lambda r: {**dic, **eval("{0:" + r.replace(' ', ',') + "}")}
    df_iter = pd.read_csv(file, names='r', iterator=True)
    yield pd.DataFrame(df_iter.get_chunk(batch_size)['r'].map(fn).values.tolist())



def tfrecord_read(filenames, feature_dict) -> tf.data.Dataset:
    """

    :param filenames:
    :param feature_dict:
        feature_dict = {
            'id': tf.io.FixedLenFeature((), tf.int64),
            'feature': tf.io.FixedLenFeature((), tf.int64),
        }
    :return:
    """
    parser_fn = partial(tf.io.parse_single_example, features=feature_dict)
    ds = (
        tf.data.TFRecordDataset(filenames, num_parallel_reads=tf.data.experimental.AUTOTUNE)
            .map(parser_fn)
    )
    return ds





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : DeepFM
# @Time         : 2020/5/22 10:03 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : https://img-blog.csdn.net/2018050820141665?watermark/2/text/aHR0cHM6Ly9ibG9nLmNzZG4ubmV0L3NvbmdiaW54dQ==/font/5a6L5L2T/fontsize/400/fill/I0JBQkFCMA==/dissolve/70
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

from tensorflow.python.feature_column.feature_column_v2 import FeatureColumn

from inn.models.BaseModel import BaseModel

import tensorflow as tf
from inn.features.fc import *

from inn.layers.DNN import DNN
from inn.layers.interaction import FM

numeric_column = tf.feature_column.numeric_column
categorical_column_with_identity = tf.feature_column.categorical_column_with_identity

nums = ['num1', 'num2']
cats = ['cat1', 'cat2']
multi_cats = ['mcat1', 'mcat2']

feature_columns = []
ncs = [numeric_column(i) for i in nums]
ccs = [categorical_column_with_identity(i, 10) for i in cats]


def fc2input(fcs: List[FeatureColumn]):
    for fc in fcs:
        yield tf.keras.layers.Input(shape=fc.shape, name=fc.name, dtype=fc.dtype)

list(fc2input(ncs))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : utils
# @Time         : 2020/4/13 6:49 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
from collections import OrderedDict
from enum import Enum


# from collections import namedtuple
# FeatType = namedtuple('FeatType', ['Cat', 'Num'])('Cat', 'Num')
class FeatureType(str, Enum):
    DenseFeature = "DenseFeature"
    SparseFeature = "SparseFeature"
    SequenceFeature = "SequenceFeature"


################################################################

import tensorflow as tf
from tensorflow.python.feature_column.feature_column_v2 import \
    FeatureColumn, NumericColumn, CategoricalColumn, SequenceCategoricalColumn, EmbeddingColumn, \
    BucketizedColumn

from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union
from tensorflow.keras.layers import Input


# tf.feature_column.numeric_column
# from tensorflow.feature_column import sequence_categorical_column_with_identity


def feature_columns_to_feature_layer_inputs(feature_columns: List[FeatureColumn], prefix=''):
    feature_layer_inputs = OrderedDict()

    for fc in feature_columns:
        if isinstance(fc, NumericColumn):
            feature_layer_inputs[fc.name] = Input(shape=fc.shape, name=prefix + fc.name, dtype=fc.dtype)
        elif isinstance(fc, CategoricalColumn):
            feature_layer_inputs[fc.name] = Input(shape=(1,), name=prefix + fc.name, dtype=fc.dtype)

        elif isinstance(fc, SequenceCategoricalColumn):  # todo: 序列特征、带权序列特征
            feature_layer_inputs[fc.name] = Input(shape=(fc.maxlen,), name=prefix + fc.name, dtype=fc.dtype)
            if fc.weight_name is not None:
                feature_layer_inputs[fc.weight_name] = Input(shape=(fc.maxlen, 1), name=prefix + fc.weight_name,
                                                             dtype="float32")
            if fc.length_name is not None:
                feature_layer_inputs[fc.length_name] = Input((1,), name=prefix + fc.length_name, dtype='int32')


        else:
            raise TypeError("Invalid feature column type,got", type(fc))

    return feature_layer_inputs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : DNN
# @Time         : 2020/5/19 10:55 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import tensorflow as tf
from inn.models import BaseModel
from inn.layers import DNN, Prediction
from inn.layers.inputs.utils import feature_columns_to_feature_layer_inputs

from tensorflow.python.feature_column.feature_column_v2 import \
    FeatureColumn, NumericColumn, CategoricalColumn, SequenceCategoricalColumn, EmbeddingColumn, \
    BucketizedColumn


class Baseline(BaseModel):

    def __init__(self, feature_columns: List[FeatureColumn], **kwargs):
        super().__init__(feature_columns=feature_columns, **kwargs)  # 覆盖基础属性

    @property
    def model(self):
        feature_layer_inputs = feature_columns_to_feature_layer_inputs(self.feature_columns)
        feature_layer_outputs = tf.keras.layers.DenseFeatures(self.feature_columns)(feature_layer_inputs)

        x = DNN()(feature_layer_outputs)

        output = Prediction(task=self.task, num_class=self.num_class)(x)

        model = tf.keras.models.Model(inputs=feature_layer_inputs, outputs=output)
        model.summary()
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : embedding
# @Time         : 2020/5/21 8:16 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from inn.utils.seed_utils import init_seeds

init_seeds(666)

import numpy as np
import tensorflow as tf

# 单值离散型
tf.keras.layers.Input(shape=(1,))
cat_feats = np.array([1, 2])
cat_feats = np.array([[1], [2]])
_ = tf.keras.layers.Embedding(10, 4, mask_zero=True)(cat_feats)
_ = tf.keras.layers.Flatten()(_)

assert _.ndim == 2

# 多值离散型: TODO: [[[1,2], [3, 4]], [[1,2], [3, 4]]] 序列的序列
tf.keras.layers.Input(shape=(2,))

seq_cat_feats = np.array([(1, 2), (2, 1)])
_ = tf.keras.layers.Embedding(10, 4, mask_zero=True)(seq_cat_feats)
_ = tf.keras.layers.GlobalAveragePooling1D()(_)

assert _.ndim == 2

from inn.features.fc import *

# embedding outputs
fc = CategoricalColumn()


inputs = []

def get_embedding_outputs(fc=CategoricalColumn()):
    _ = columns2inputs(fc)

    if isinstance(fc, CategoricalColumn):
        return tf.keras.layers.Embedding(fc.vocabulary_size, fc.embedding_dim, mask_zero=True)(_)

    elif isinstance(fc, SequenceCategoricalColumn):
        _ = tf.keras.layers.Embedding(fc.vocabulary_size, fc.embedding_dim, mask_zero=True)(_)
        if fc.combiner == 'mean':
            return tf.keras.layers.GlobalAveragePooling1D()(_)
        elif fc.combiner == 'max':
            return tf.keras.layers.GlobalMaxPooling1D()(_)
        else:
            raise ValueError("Unknown combiner")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : tensor_utils
# @Time         : 2020/5/20 10:59 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

import numpy as np
import tensorflow as tf


def arr2sparse(arr):
    arr_tensor = tf.constant(np.array(arr))
    arr_idx = tf.where(tf.not_equal(arr_tensor, 0))
    arr_sparse = tf.SparseTensor(arr_idx, tf.gather_nd(arr_tensor, arr_idx), arr_tensor.get_shape())
    return arr_sparse


def sparse2dense(tensor):
    return tf.sparse.to_dense(tensor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : utils
# @Time         : 2020/4/10 10:58 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import tensorflow as tf


def get_activation_by_num_class(num_class=2, task2activation=None):
    task2activation = {'regression': None, 'binary': 'sigmoid', 'multiclass': 'softmax'}
    if task2activation:
        task2activation.update(task2activation)

    if num_class == 1:
        return task2activation['regression']
    elif num_class == 2:
        return task2activation['binary']
    elif num_class >= 3:
        return task2activation['multiclass']
    else:
        raise ValueError(f"Unsupported {num_class}")


# tf.divide(x, y, name=name)
# tf.nn.softmax(logits, axis=dim, name=name)
# tf.reduce_mean(input_tensor, axis=None, keepdims=False, name=None)
# tf.reduce_sum(input_tensor, axis=None, keepdims=False, name=None)
# tf.reduce_max(input_tensor, axis=None, keepdims=False, name=None)


class Hash(tf.keras.layers.Layer):
    """
    hash the input to [0,num_buckets)
    if mask_zero = True,0 or 0.0 will be set to 0,other value will be set in range[1,num_buckets)
    """

    def __init__(self, num_buckets, mask_zero=False, **kwargs):
        self.num_buckets = num_buckets
        self.mask_zero = mask_zero
        super(Hash, self).__init__(**kwargs)

    def build(self, input_shape):
        # Be sure to call this somewhere!
        super(Hash, self).build(input_shape)

    def call(self, x, mask=None, **kwargs):
        if x.dtype != tf.string:
            x = tf.as_string(x, )

        hash_x = tf.strings.to_hash_bucket_fast(x, self.num_buckets if not self.mask_zero else self.num_buckets - 1,
                                                name=None)  # weak hash
        if self.mask_zero:
            mask_1 = tf.cast(tf.not_equal(x, "0"), 'int64')
            mask_2 = tf.cast(tf.not_equal(x, "0.0"), 'int64')
            mask = mask_1 * mask_2
            hash_x = (hash_x + 1) * mask
        return hash_x

    def compute_mask(self, inputs, mask=None):
        return None

    def get_config(self, ):
        config = {'num_buckets': self.num_buckets, 'mask_zero': self.mask_zero}
        base_config = super(Hash, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-10-21 13:00
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import pandas as pd
from sklearn.model_selection import train_test_split
from .Dataset import Dataset


def load_iris(batch_size=128):
    from sklearn import datasets
    ds = Dataset(batch_size)
    X, y = datasets.load_iris(True)
    return ds.from_cache(X, y)


def load_heart(return_ds=True):
    df = pd.read_csv('https://storage.googleapis.com/applied-dl/heart.csv')
    if return_ds:
        train, test = train_test_split(df, test_size=0.2)
        ds_train = Dataset().from_cache(train.drop('target', 1), train['target'])
        ds_test = Dataset().from_cache(test.drop('target', 1), test['target'])
        return ds_train, ds_test
    else:
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : logger
# @Time         : 2020-04-10 16:35
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import os
from loguru import logger
from pathlib import Path

LOG_PATH = os.environ.get("LOG_PATH")
if LOG_PATH:
    logger.add(Path(LOG_PATH),
               rotation="100 MB",
               retention='1 days',
               enqueue=True,  # 异步
               encoding="utf-8",
               backtrace=True,
               diagnose=True,
               level="INFO")

# import os
# import sys
# from loguru import logger
#
# BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#
# log_file_path = os.path.join(BASE_DIR, 'Log/my.log')
# err_log_file_path = os.path.join(BASE_DIR, 'Log/err.log')
#
# logger.add(sys.stderr, format="{time} {level} {message}", filter="my_module", level="INFO")
# # logger.add(s)
# logger.add(log_file_path, rotation="500 MB", encoding='utf-8')  # Automatically rotate too big file
# logger.add(err_log_file_path, rotation="500 MB", encoding='utf-8', level='ERROR')  # Automatically rotate too big file
# logger.debug("That's it, beautiful and simple logging!")
# logger.debug("中文日志可以不")
# logger.error("严重错误")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : Prediction
# @Time         : 2020/5/19 6:14 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf
from .utils import get_activation_by_num_class


class Prediction(tf.keras.layers.Layer):

    def __init__(self, num_class=2, name='Prediction', **kwargs):
        super().__init__(name=name, **kwargs)

        self.num_class = num_class
        self.activation = get_activation_by_num_class(num_class)

    def build(self, input_shape):
        super().build(input_shape)
        units = self.num_class if self.num_class > 2 else 1  # 多分类输出 num_class 维
        self.fc = tf.keras.layers.Dense(units, activation=self.activation)

    def call(self, inputs, **kwargs):
        return self.fc(inputs)

    def compute_output_shape(self, input_shape):
        if self.task == "multiclass":
            return (None, self.num_class)
        else:
            return (None, 1)

    def get_config(self, ):
        base_config = super().get_config()
        config = {'task': self.task, 'num_class': self.num_class}
        return {**base_config, **config}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-09-11 12:02
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf
from tensorflow.keras.utils import get_custom_objects

from sklearn.metrics import roc_auc_score


# # 注册函数
# funcs = [tf.keras.metrics.AUC()]
# custom_objects = {func.__name__: Activation(func) for func in funcs}
#
# get_custom_objects().update()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : tfrecord_write
# @Time         : 2020/4/22 1:29 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import tensorflow as tf


def _float_list(value):
    if not isinstance(value, list):
        value = [value]

    float_list = tf.train.FloatList(value=value)
    return tf.train.Feature(float_list=float_list)


def tf_train_example(feature=None):
    if feature is None:
        feature = {}
    _ = tf.train.Example(features=tf.train.Features(feature=feature))
    return _


with tf.io.TFRecordWriter('./tf.record') as f:
    features = tf.train.Features(
        feature={
            'num': _float_list(1),
            'seq': _float_list([1, 2, 3]),
            'label': _float_list(0)
        })
    eaxmple = tf.train.Example(features=features)
    f.write(eaxmple.SerializeToString())


def parser_fn(example_photo):
    features = {
        'num': tf.io.FixedLenFeature((), tf.float32),
        'seq': tf.io.FixedLenSequenceFeature((), tf.float32, allow_missing=True),
        'label': tf.io.FixedLenFeature((), tf.float32),

    }
    parsed_features = tf.io.parse_single_example(example_photo, features=features)
    return parsed_features


data = (
    tf.data.TFRecordDataset('./tf.record')
        .map(parser_fn)
)

print(list(data.as_numpy_iterator()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : cache_utils
# @Time         : 2020/4/24 1:28 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : seq_padding
# @Time         : 2020/5/18 4:34 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


from tensorflow.keras.preprocessing.text import Tokenizer
from tensorflow.keras.preprocessing.sequence import pad_sequences


class Sequence(object):
    """doc
    词袋模型：我们可以为数据集中的所有单词制作一张词表，然后将每个单词和一个唯一的索引关联。
    每个句子都是由一串数字组成，这串数字是词表中的独立单词对应的个数。
    通过列表中的索引，我们可以统计出句子中某个单词出现的次数。
    """

    def __init__(self, num_words=20000, maxlen=None):
        """
        OOV对应的索引是1
        :param maxlen: 句子序列最大长度
        :param num_words: top num_words-1(词频降序)：保留最常见的num_words-1词
        """
        self.maxlen = maxlen
        self._num_words = num_words
        self.tokenizer = None
        self.word_index = None

    def fit(self, docs):
        """
        :param corpus: ['some thing to do', 'some thing to drink']与sklearn提取文本特征一致
        """
        print('Create Bag Of Words ...')
        self.tokenizer = Tokenizer(self._num_words, lower=False, oov_token='OOV')  # 不改变大小写（需提前预处理）
        self.tokenizer.fit_on_texts(docs)
        self.word_index = self.tokenizer.word_index
        print("Get Unique Words In Corpus: %s" % len(self.word_index))
        return self

    def transform(self, docs):
        print('Docs To Sequences ...')
        sequences = self.tokenizer.texts_to_sequences(docs)
        # 补后面 截断后面
        pad_docs = pad_sequences(sequences, self.maxlen, padding='post', truncating='post')
        if self.maxlen is None:
            self.maxlen = pad_docs.shape[1]
        return pad_docs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : backend
# @Time         : 2019-09-10 15:16
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf

#
# from tensorflow.python import keras
#
# utils = keras.utils
# activations = keras.activations
# applications = keras.applications
# backend = keras.backend
# datasets = keras.datasets
# engine = keras.engine
# layers = keras.layers
# preprocessing = keras.preprocessing
# wrappers = keras.wrappers
# callbacks = keras.callbacks
# constraints = keras.constraints
# initializers = keras.initializers
# metrics = keras.metrics
# models = keras.models
# losses = keras.losses
# optimizers = keras.optimizers
# regularizers = keras.regularizers

custom_objects = {
    # metrics
    'auc': tf.keras.metrics.AUC

    # losses

    # optimizers

    # callbacks

    # layers

    # activations

    # 'Embedding': Embedding,
    # 'BiasAdd': BiasAdd,
    # 'MultiHeadAttention': MultiHeadAttention,
    # 'LayerNormalization': LayerNormalization,
    # 'PositionEmbedding': PositionEmbedding,
    # 'RelativePositionEmbedding': RelativePositionEmbedding,
    # 'RelativePositionEmbeddingT5': RelativePositionEmbeddingT5,
    # 'FeedForward': FeedForward,
    # 'ConditionalRandomField': ConditionalRandomField,
    # 'MaximumEntropyMarkovModel': MaximumEntropyMarkovModel,
    # 'Loss': Loss,
}

tf.keras.utils.get_custom_objects().update(custom_objects)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : seed_utils
# @Time         : 2020/4/21 1:15 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import os
import random as rn
import numpy as np
import tensorflow as tf


def init_seeds(seed=666):
    os.environ['PYTHONHASHSEED'] = '0'
    os.environ['CUDA_VISIBLE_DEVICES'] = '0'

    # The below is necessary for starting Numpy generated random numbers
    # in a well-defined initial state.

    np.random.seed(seed)

    # The below is necessary for starting core Python generated random numbers
    # in a well-defined state.

    rn.seed(seed)

    tf.compat.v1.reset_default_graph()
    tf.random.set_seed(seed)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'CountEncoder'
__author__ = 'JieYuan'
__mtime__ = '19-1-10'
"""
from collections import OrderedDict

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class CountEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, dropna=False, normalize=False):
        """

        :param dropna: 缺失值是否计数，默认计数
        :param normalize: 频数还是频率，默认频数
        """
        self.dropna = dropna
        self.normalize = normalize
        self.mapper = None

    def fit(self, y):
        self.mapper = (pd.Series(y).value_counts(self.normalize, dropna=self.dropna)
                       .to_dict(OrderedDict))
        return self

    def transform(self, y):
        """不在训练集的补0，不经常出现补0"""
        return pd.Series(y).map(self.mapper).fillna(0)


if __name__ == '__main__':
    import numpy as np

    s = ['a', 'a', 'a', 'b', 'b'] + [np.nan] * 10
    ce = CountEncoder()
    print(ce.fit_transform(s))
    print(ce.mapper)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : df_utils
# @Time         : 2020/5/18 4:42 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import re


def df_select_columns(pattern, columns):
    return [c for c in columns if re.search(pattern, c)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : config_utils
# @Time         : 2020/5/8 4:13 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


from pyapollo import ApolloClient


class Config(object):
    """
    http://106.54.227.205/
    https://github.com/ctripcorp/apollo

    client = ApolloClient('400000008', 'c3', 'http://10.114.38.22:8080') # port不对
    """

    def __init__(self, ):
        pass

    def apollo_client(self, app_id='yuanjie', cluster='yuanjie', config_server_url="http://106.54.227.205:8080"):
        """
            client.get_value('yuanjie')
        :param app_id:
        :param cluster:
        :param config_server_url:
        :return:
        """
        client = ApolloClient(app_id, cluster, config_server_url)
        client.start()
        return client



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : shake
# @Time         : 2020/4/27 11:09 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


# @jit(forceobj=True)
# def get_scores():
#     s = []
#     for n in tqdm_notebook(range(1300, 1800, 1)):
#         s.append(roc_auc_score([1]*n+[0]*(170610-n), [1]*1+[0]*(170610-1)))
#     return s


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : __init__.py
# @Time         : 2020/5/20 10:37 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : Add
# @Time         : 2020/4/10 11:20 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import tensorflow as tf


class Add(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        # Be sure to call this somewhere!
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        if not isinstance(inputs, list):
            return inputs
        if len(inputs) == 1:
            return inputs[0]
        if len(inputs) == 0:
            return tf.constant([[0.0]])

        return tf.keras.layers.add(inputs) # 为了通用



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : BaseModel
# @Time         : 2020/5/19 2:24 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

from collections import namedtuple, OrderedDict

from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import tensorflow as tf

from ..features.fc import *
from ..features.fc import utils as fc_utils

from inn.layers import Prediction


class BaseModel(object):

    def __init__(self,
                 fcs: List[Column] = None,
                 num_class=2,
                 early_stopping_epochs=3,
                 model_name='BaseModel'):
        self.fcs = fcs
        self.fc2input = list(fc_utils.get_fc2input_map(fcs))
        self.inputs = [v for k, v in self.fc2input]

        self.ncs = [fc for fc, input in self.fc2input if isinstance(fc, NumericColumn)]
        self.ccs = [fc for fc, input in self.fc2input if isinstance(fc, CategoricalColumn)]
        self.sccs = [fc for fc, input in self.fc2input if isinstance(fc, SequenceCategoricalColumn)]

        self.num_class = num_class
        # self.activation = get_activation_by_num_class(num_class)

        self.early_stopping_epochs = early_stopping_epochs
        self.model_name = model_name

        #################################
        self.predict_layer = Prediction(num_class)
        self.model = self._build_model()

    @abstractmethod
    def _build_model(self, **kwargs):
        raise NotImplementedError()

    def fit(self, *args, **kwargs):
        return self.model.fit(*args, **kwargs)

    def predict(self, X):
        return self.model.predict(X)

    def summary(self, return_summary=True, to_file='model.png',
                show_shapes=True,
                rankdir='TB',
                dpi=96):
        if return_summary:
            self.model.summary()

        tf.keras.utils.plot_model(self.model, to_file=to_file,
                                  show_shapes=show_shapes,
                                  rankdir=rankdir,
                                  dpi=dpi)

    def compile(self,
                optimizer='rmsprop',
                loss=None,
                metrics=None,
                loss_weights=None,
                sample_weight_mode=None,
                weighted_metrics=None):
        self.model.compile(optimizer=optimizer,
                           loss=loss,
                           metrics=metrics,
                           loss_weights=loss_weights,
                           sample_weight_mode=sample_weight_mode,
                           weighted_metrics=weighted_metrics)

    def callbacks(self):
        callbacks_list = [
            tf.keras.callbacks.ReduceLROnPlateau(factor=0.9, patience=2, verbose=1, min_lr=0.0001),
            # annealer = LearningRateScheduler(lambda x: min(0.01 * 0.9 ** x, 0.001), verbose=1)

            tf.keras.callbacks.ModelCheckpoint("filepath",
                                               monitor='val_loss',
                                               verbose=0,
                                               save_best_only=True,
                                               save_weights_only=False,
                                               mode='auto',
                                               save_freq='epoch'),

            tf.keras.callbacks.EarlyStopping(monitor='val_loss',
                                             min_delta=0,
                                             patience=self.early_stopping_epochs,
                                             verbose=0,
                                             mode='auto',
                                             baseline=None,
                                             restore_best_weights=False)
        ]
        return callbacks_list

    # todo: 学习率clr_callback, WandbCallback



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : FM_
# @Time         : 2020/4/21 1:43 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import torch
import torch.nn as nn


class FM(nn.Module):
    """Factorization Machine models pairwise (order-2) feature interactions
     without linear term and bias.
      Input shape
        - 3D tensor with shape: ``(batch_size,field_size,embedding_size)``.
      Output shape
        - 2D tensor with shape: ``(batch_size, 1)``.
      References
        - [Factorization Machines](https://www.csie.ntu.edu.tw/~b97053/paper/Rendle2010FM.pdf)
    """

    def __init__(self):
        super(FM, self).__init__()

    def forward(self, inputs):
        fm_input = inputs

        square_of_sum = torch.pow(torch.sum(fm_input, dim=1, keepdim=True), 2)
        sum_of_square = torch.sum(fm_input * fm_input, dim=1, keepdim=True)
        cross_term = square_of_sum - sum_of_square
        cross_term = 0.5 * torch.sum(cross_term, dim=2, keepdim=False)

        return cross_term



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions and classes related to optimization (weight updates)."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import tensorflow as tf





class AdamWeightDecayOptimizer(tf.train.Optimizer):
    """A basic Adam optimizer that includes "correct" L2 weight decay."""

    def __init__(self,
                 learning_rate,
                 weight_decay_rate=0.0,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=1e-6,
                 exclude_from_weight_decay=None,
                 name="AdamWeightDecayOptimizer"):
        """Constructs a AdamWeightDecayOptimizer."""
        super(AdamWeightDecayOptimizer, self).__init__(False, name)

        self.learning_rate = learning_rate
        self.weight_decay_rate = weight_decay_rate
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.exclude_from_weight_decay = exclude_from_weight_decay

    def apply_gradients(self, grads_and_vars, global_step=None, name=None):
        """See base class."""
        assignments = []
        for (grad, param) in grads_and_vars:
            if grad is None or param is None:
                continue

            param_name = self._get_variable_name(param.name)

            m = tf.get_variable(
                name=param_name + "/adam_m",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.zeros_initializer())
            v = tf.get_variable(
                name=param_name + "/adam_v",
                shape=param.shape.as_list(),
                dtype=tf.float32,
                trainable=False,
                initializer=tf.zeros_initializer())

            # Standard Adam update.
            next_m = (
                    tf.multiply(self.beta_1, m) + tf.multiply(1.0 - self.beta_1, grad))
            next_v = (
                    tf.multiply(self.beta_2, v) + tf.multiply(1.0 - self.beta_2,
                                                              tf.square(grad)))

            update = next_m / (tf.sqrt(next_v) + self.epsilon)

            # Just adding the square of the weights to the loss function is *not*
            # the correct way of using L2 regularization/weight decay with Adam,
            # since that will interact with the m and v parameters in strange ways.
            #
            # Instead we want ot decay the weights in a manner that doesn't interact
            # with the m/v parameters. This is equivalent to adding the square
            # of the weights to the loss with plain (non-momentum) SGD.
            if self._do_use_weight_decay(param_name):
                update += self.weight_decay_rate * param

            update_with_lr = self.learning_rate * update

            next_param = param - update_with_lr

            assignments.extend(
                [param.assign(next_param),
                 m.assign(next_m),
                 v.assign(next_v)])
        return tf.group(*assignments, name=name)

    def _do_use_weight_decay(self, param_name):
        """Whether to use L2 weight decay for `param_name`."""
        if not self.weight_decay_rate:
            return False
        if self.exclude_from_weight_decay:
            for r in self.exclude_from_weight_decay:
                if re.search(r, param_name) is not None:
                    return False
        return True

    def _get_variable_name(self, param_name):
        """Get the variable name from the tensor name."""
        m = re.match("^(.*):\\d+$", param_name)
        if m is not None:
            param_name = m.group(1)
        return param_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : feature_column
# @Time         : 2020/5/20 10:38 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

from collections import OrderedDict

import tensorflow as tf
from tensorflow.keras.layers import Input

from tensorflow.python.feature_column.feature_column_v2 import \
    FeatureColumn, NumericColumn, \
    CategoricalColumn, SequenceCategoricalColumn, BucketizedColumn, \
    IndicatorColumn, EmbeddingColumn, IdentityCategoricalColumn


def feature_columns_to_feature_layer_inputs(feature_columns: List[FeatureColumn]):
    feature_layer_inputs = OrderedDict()

    for fc in feature_columns:
        # NumericColumn: BucketizedColumn 离散化
        if isinstance(fc, NumericColumn):
            feature_layer_inputs[fc.name] = Input(shape=fc.shape, name=fc.name, dtype=fc.dtype)

        # CategoricalColumn: IndicatorColumn + EmbeddingColumn
        elif isinstance(fc, IndicatorColumn):

            if isinstance(fc.categorical_column, IdentityCategoricalColumn):
                dtype = tf.int64
            else:
                dtype = fc.categorical_column.dtype

            feature_layer_inputs[fc.name[:-10]] = Input(shape=(1,), name=fc.name[:-10], dtype=dtype)


        elif isinstance(fc, CategoricalColumn):
            feature_layer_inputs[fc.name] = Input(shape=(1,), name=fc.name, dtype=fc.dtype)

        elif isinstance(fc, SequenceCategoricalColumn):  # todo: 序列特征、带权序列特征
            feature_layer_inputs[fc.name] = Input(shape=(fc.maxlen,), name=fc.name, dtype=fc.dtype)
            if fc.weight_name is not None:
                feature_layer_inputs[fc.weight_name] = Input(shape=(fc.maxlen, 1), name=fc.weight_name,
                                                             dtype="float32")
            if fc.length_name is not None:
                feature_layer_inputs[fc.length_name] = Input((1,), name=fc.length_name, dtype='int32')


        elif isinstance(fc, (BucketizedColumn, EmbeddingColumn)):  # 有些特征列是建立在其他输入之上的，不需要输入
            pass

        else:
            raise TypeError("Invalid feature column type,got", type(fc))

    return feature_layer_inputs


# EmbeddingColumn
def embedding_column(feature_columns: List[FeatureColumn], get_dim=lambda nunique: int(6 * nunique ** 0.25)):
    ecs = []
    for fc in feature_columns:
        assert isinstance(fc,
                          CategoricalColumn), f"{fc} object has no attribute 'num_buckets'"  # hasattr(fc, 'num_buckets')

        dim = get_dim(fc.num_buckets + 1) if callable(get_dim) else get_dim  # oov
        ec = tf.feature_column.embedding_column(fc, dim)
        ecs.append(ec)

    return ecs

#################################

tf.feature_column.numeric_column
tf.feature_column.categorical_column_with_identity
tf.feature_column.categorical_column_with_hash_bucket



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : examples
# @Time         : 2020/4/20 8:21 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : https://mp.weixin.qq.com/s/5jfOahNKnUjTre0O2655IA


import tensorflow as tf
from tensorflow.python.feature_column import feature_column_lib
# tf.sparse.to_dense()



def categorical_list_column():
    column = tf.feature_column.categorical_column_with_vocabulary_list(
        # 特征列的名称
        key="feature",
        # 有效取值列表，列表的下标对应转换的数值。即，value1会被映射为0
        vocabulary_list=["value1", "value2", "value3"],
        # 取值的类型，只支持string和integer，这个会根据vocabulary_list自动推断出来
        dtype=tf.string,
        # 当取值不在vocabulary_list中时，会被映射的数值，默认为-1
        # 当该值不为-1时，num_oov_buckets必须设置为0。即两者不能同时起作用
        default_value=-1,
        # 作用同default_value，但是两者不能同时起作用。
        # 将超出的取值映射到[len(vocabulary), len(vocabulary) + num_oov_buckets)内
        # 默认取值为0
        # 当该值不为0时，default_value必须设置为-1
        # 当default_value和num_oov_buckets都取默认值时，会被映射为-1
        num_oov_buckets=3)
    feature_cache = feature_column_lib.FeatureTransformationCache(features={
        # feature对应的值可以为Tensor，也可以为SparseTensor
        "feature": tf.constant(value=[
            [["value1", "value2"], ["value3", "value3"]],
            [["value3", "value5"], ["value4", "value4"]]
        ])
    })
    # IdWeightPair(id_tensor, weight_tensor)
    return column.get_sparse_tensors(transformation_cache=feature_cache, state_manager=None)


if __name__ == '__main__':
    c = categorical_list_column()
    print(c.id_tensor.dtype)
    print(c.id_tensor.shape)
    print(c.id_tensor.dense_shape)
    print(c.id_tensor.indices)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : FM_1
# @Time         : 2020/4/21 11:35 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : /Users/yuanjie/Desktop/Projects/Python/Awesome-RecSystem-Models/Model/FM_TensorFlow.py





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : Multihot
# @Time         : 2020/5/20 10:42 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from sklearn.feature_extraction.text import CountVectorizer


class MultiHot(CountVectorizer):

    def __init__(self, **kwargs):
        super().__init__(tokenizer=lambda x: x, lowercase=False, **kwargs)


import tensorflow as tf
from tensorflow.python.feature_column import feature_column_lib


def multihot(array=None, num_buckets=999, default_value=0):
    categorical_column = tf.feature_column.categorical_column_with_identity(
        key="feature",
        num_buckets=num_buckets,
        default_value=default_value)

    # multi-hot编码，出现多次的相同值，会累加
    column = tf.feature_column.indicator_column(categorical_column=categorical_column)

    if array is None:
        array = [[1, 2, 1],
                 [3, 4, 5]]
    feature_cache = feature_column_lib.FeatureTransformationCache(features={"feature": tf.constant(value=array)})
    _ = column.get_dense_tensor(transformation_cache=feature_cache, state_manager=None)

    # _ = tf.keras.layers.DenseFeatures([column])({'feature': array})
    return _



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : __init__.py
# @Time         : 2020/4/21 1:39 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import os

if os.environ.get('NN') == 'torch':  # torch/tensorflow
    from .FM_ import FM
else:
    from .FM import FM



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-09-10 15:22
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf

from .Add import Add
from .NoMask import NoMask
from .Linear import Linear
from .DNN import DNN
from .Prediction import Prediction

from tensorflow.keras.layers import *


def add_func(inputs):
    return Add()(inputs)


def concat_func(inputs, axis=-1, mask=False):
    if not mask:
        inputs = list(map(NoMask(), inputs))
    if len(inputs) == 1:
        return inputs[0]
    else:
        return tf.keras.layers.Concatenate(axis=axis)(inputs)


# from functools import partial
# from collections import Iterable
#
# from keras.layers import Dense
#
# # last layer: 'softmax' or 'sigmoid'
# def getDenseList(unitsList, activation=None):
#     assert isinstance(unitsList, Iterable)
#     dense = partial(Dense, activation=activation)
#     return map(dense, unitsList)


# 通过定义个操作 Tensor 的函数，然后将其添加到 keras 系统中即可。
custom_objects = {
    # 'Embedding': Embedding,
    # 'BiasAdd': BiasAdd,
    # 'MultiHeadAttention': MultiHeadAttention,
    # 'LayerNormalization': LayerNormalization,
    # 'PositionEmbedding': PositionEmbedding,
    # 'RelativePositionEmbedding': RelativePositionEmbedding,
    # 'RelativePositionEmbeddingT5': RelativePositionEmbeddingT5,
    # 'FeedForward': FeedForward,
    # 'ConditionalRandomField': ConditionalRandomField,
    # 'MaximumEntropyMarkovModel': MaximumEntropyMarkovModel,
    # 'Loss': Loss,
}

tf.keras.utils.get_custom_objects().update(custom_objects)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : scheduler.py
# @Time         : 2020/5/18 2:28 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import os

os.system("pip install -U --no-cache-dir -i https://pypi.tuna.tsinghua.edu.cn/simple loguru schedule")
import time
import schedule

from loguru import logger

logger.add('./scheduler.log',
           rotation="100 MB",
           enqueue=True,  # 异步
           encoding="utf-8",
           backtrace=True,
           diagnose=True,
           level="INFO")


def job():
    logger.info("I'm working...")
    from mi.db import Hive
    Hive()

    for _ in range(11):
        os.system("python /fds/seize.py")


#     _ = os.popen("ps aux | grep seize.py | awk '{print $2}'").read().split()
#     [os.system(f"kill -9 {i}") for i in _]


schedule.every().day.at("00:00").do(job)

while True:
    schedule.run_pending()
    time.sleep(1)



# -*- coding: utf-8 -*-
"""
__title__ = 'z_score'
__author__ = 'JieYuan'
__mtime__ = '2018/7/31'
"""
import numpy as np
import pandas as pd
from sklearn.ensemble import IsolationForest


class OutlierEncoder(object):
    """如何判定和处理异常值，需要结合实际。
    若数据服从正太分布，在3σ原则下，异常值被定义为一组测定值中与平均值的偏差超过3倍标准差的值，
    因为在正态分布的假设下，距离平均值3σ之外的值出现的概率小于0.003
    1.删除含有异常值的记录
    2.将异常值视为缺失值，交给缺失值处理方法来处理
    3.用中位数来修正
    4.不处理
    """

    def __init__(self, mode='isf'):
        """
        :param mode: 'isf', 'box_plot', 'zscore_median'
        """
        self.mode = mode

    def transform(self, x: pd.Series, return_bool=True):
        """返回异常值的索引或者0/1"""
        if return_bool:
            return self.__getattribute__(self.mode)(x)
        else:
            return self.__getattribute__(self.mode)(x).astype(int)

    def get_outlier_idx(self, x: pd.Series):
        """投票"""
        _ = [self.isf(x), self.box_plot(x), self.zscore_median(x)]
        _ = pd.DataFrame(_).sum() > 1
        return np.where(_)[0].tolist()

    def isf(self, x):
        return IsolationForest(behaviour='new', contamination='auto', n_jobs=-1).fit_predict(x.to_frame()) == -1

    def box_plot(self, x):
        bound_func = lambda a, b: 2.5 * a - 1.5 * b
        q1, q3 = x.quantile(0.25), x.quantile(0.75)
        lower_bound, upper_bound = bound_func(q1, q3), bound_func(q3, q1)
        return ~x.between(lower_bound, upper_bound)

    def zscore_median(self, x, threshold=3):
        """3 sigma"""
        x_median = (x - x.median()).abs()
        MAD = x_median.median()
        zscore = 0.6475 * x_median / MAD
        return zscore > threshold

    # def zscore_mean(self, x, threshold=3):
    #     zscore = (x - x.mean()) / x.std()
    #     return zscore.abs() > threshold


if __name__ == '__main__':
    s = pd.Series([-10, 1, 2, 3, 4, 5, 100])
    print(s)

    outlier = OutlierEncoder('zscore_median')
    print(outlier.get_outlier_idx(s))
    print(outlier.transform(s))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : cprint
# @Time         : 2020-02-12 00:09
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


def cprint(string='Hello World !', bg='blue', fg='', mode=1, return_str=False):
    """
    :param s: string
    :param fg/bg: foreground/background
        'black', 'red', 'green', 'yellow', 'blue', 'purple', 'cyan', 'white'
    :param mode:
        0（默认值）
        1（高亮）
        22（非粗体）
        4（下划线）
        24（非下划线）
        5（闪烁）
        25（非闪烁）
        7（反显）
        27（非反显）
        https://www.cnblogs.com/hellojesson/p/5961570.html
    :return:
    """
    colors = ['green', 'yellow', 'black', 'cyan', 'blue', 'red', 'white', 'purple']
    fc = dict(zip(colors, range(40, 97)))
    bc = dict(zip(colors, range(90, 97)))
    _ = f"\033[{mode};{fc[fg]};{bc[bg]}m{string}\033[0m" if fg else f"\033[{mode};{bc[bg]}m{string}\033[0m"
    if return_str:
        return _
    else:
        print(_)


if __name__ == '__main__':
    cprint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : Adversarialvalidation
# @Time         : 2019-09-22 21:49
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


"""
对抗性验证
https://www.kaggle.com/kevinbonnes/adversarial-validation
"""


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : baseline
# @Time         : 2020/5/19 5:33 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf
from inn.dateset import Dataset, load_heart
from inn.models import Baseline

df = load_heart(0)

df.head()

num_cols = ['age', 'sex']  # 选择部分列

ds = Dataset().from_cache(df[num_cols], df['target'], shuffle_seed=1)

model = Baseline(list(map(tf.feature_column.numeric_column, num_cols))).model

model.compile(loss='categorical_crossentropy')

tf.keras.utils.plot_model(model, show_shapes=True)

model.fit(ds, epochs=10, validation_data=ds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : pipe
# @Time         : 2020-04-10 16:41
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
from .cprint import cprint
from functools import update_wrapper
from tqdm.auto import tqdm


# 参照 tql_python
class pipe:
    """I am very like a linux pipe"""

    def __init__(self, function):
        self.function = function
        update_wrapper(self, function)

    def __ror__(self, other):
        return self.function(other)

    def __call__(self, *args, **kwargs):
        return pipe(lambda x: self.function(x, *args, **kwargs))


# print
@pipe
def xprint(obj, mode=None, bg='blue'):
    if mode:
        for i in obj:
            cprint(i)
            print('\n')
    else:
        cprint(obj, bg)


xtqdm = pipe(lambda iterable, desc=None: tqdm(iterable, desc))

# base types
xtuple, xlist, xset = pipe(tuple), pipe(list), pipe(set)



"""
auc表示（正样本概率>负样本概率）的概率 # 可以算期望

根据线上auc猜测正负样本数
1/正样本数 = 2 * auc - 1  # gini = 2 * auc - 1 与gini公式一致
如果auc<0.5就是负样本数

预测错的个数 > (1 - auc) * 正样本数 
# 至少 (1 - auc) * 正样本数 未召回
# 至多有 auc * 正样本数 召回 准确率不超过 auc
"""

"""f1
提交一个正确的正样本(确保f1不为0)得到线上f1
猜一个1其他全为0: 正样本数 = 2 / f1 - 1
猜一个0其他全为1: 正样本数 = (总样本数 - 1) / (2 / f1 - 1)  # 概率更大
"""

"""acc
猜一个1其他全为0: 正样本数 = 总样本数 * acc - 1
猜一个0其他全为1: 负样本数 = 总样本数 * acc - 1
"""

import numpy as np


def get_p_n(score, pred):
    """通过logloss得分估算正样本率"""
    a = - score - np.log(1 - pred)
    b = np.log(pred / (1 - pred))
    return a / b


def result_scaling(x=0.37, rate_train_true=0.37, rate_test_true=0.165):
    """根据log_loss缩放结果，适用于线上线下样本分布不一致
    Refer:
    https://www.kaggle.com/c/quora-question-pairs/discussion/31179
    https://www.kaggle.com/badat0202/estimate-distribution-of-data-in-lb
    """
    a = rate_test_true / rate_train_true
    b = (1 - rate_test_true) / (1 - rate_train_true)
    scale_pos_weight = a / b  # 利用scale_pos_weight缩放结果
    print("Xgb/Lgb scale_pos_weight: %s" % scale_pos_weight)
    return a * x / (a * x + b * (1 - x))


def get_weight(y):
    class_weight = dict(enumerate(len(y) / (2 * np.bincount(y))))
    sample_weight = [class_weight[i] for i in y]
    return class_weight, sample_weight



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : FM
# @Time         : 2019-07-16 16:12
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :


"""
https://github.com/jfpuget/LibFM_in_Keras/blob/master/keras_blog.ipynb
Trick:
1. 单模型
2. 提取embedding共享特征与原特征拼接 + 其他基模型
"""
import pandas as pd
from tensorflow.python.keras.models import Model
from tensorflow.python.keras.layers.normalization import BatchNormalization
from tensorflow.python.keras.layers import Input, Embedding, Dense, Flatten, Concatenate, Dot, Reshape, Add, Subtract
from tensorflow.keras.optimizers import Adam
from tensorflow.python.keras.regularizers import l2
from tensorflow.python.keras.callbacks import EarlyStopping, ModelCheckpoint


class KerasFM(object):

    def __init__(self, k_latent=2, kernel_l2=0.1):
        self.k_latent = k_latent  # TODO: embedding dim 经验值
        self.kernel_l2 = kernel_l2

    def build_model(self, f_sizes):
        """
        :param f_size: sparse feature nunique
        :return:
        """
        dim_input = len(f_sizes)  # +1

        input_x = [Input(shape=(1,)) for i in range(dim_input)]  # 多列 sparse feature
        biases = [self.get_embed(x, size, 1) for (x, size) in zip(input_x, f_sizes)]

        factors = [self.get_embed(x, size) for (x, size) in zip(input_x, f_sizes)]

        s = Add()(factors)
        diffs = [Subtract()([s, x]) for x in factors]
        dots = [Dot(axes=1)([d, x]) for d, x in zip(diffs, factors)]

        x = Concatenate()(biases + dots)
        x = BatchNormalization()(x)
        output = Dense(1, activation='relu', kernel_regularizer=l2(self.kernel_l2))(x)
        model = Model(inputs=input_x, outputs=[output])
        model.compile(optimizer=Adam(clipnorm=0.5), loss='mean_squared_error')  # TODO: radam

        output_f = factors + biases
        model_features = Model(inputs=input_x, outputs=output_f)
        return model, model_features

    def get_embed(self, x_input, x_size, embedding_l2=0.0002):
        if x_size > 0:  # category
            embed = Embedding(x_size, self.k_latent, input_length=1, embeddings_regularizer=l2(embedding_l2))(x_input)
            embed = Flatten()(embed)
        else:
            embed = Dense(self.k_latent, kernel_regularizer=l2(embedding_l2))(x_input)
        return embed

    def get_factors_biases(self, Xs, feature_names, model_features):
        """We can now retrieve the factors and the biases"""
        X_pred = model_features.predict(Xs, 2 ** 10)
        n = len(X_pred) // 2
        factors = X_pred[:n]
        biases = X_pred[-n:]

        df = pd.DataFrame()
        for f, X_p in zip(feature_names, factors):
            for i in range(self.k_latent):
                df[f'{f}_fm_factor_{i}'] = X_p[:, i]

        for f, X_p in zip(feature_names, biases):
            df[f'{f}_fm_bias'] = X_p[:, 0]
        return df


if __name__ == '__main__':
    from sklearn.datasets import make_classification
    from sklearn.preprocessing import LabelEncoder

    X, y = make_classification(n_samples=1000, n_features=7, random_state=42)

    Xs = []
    f_sizes = []
    for i in range(X.shape[1]):
        X[:, i] = LabelEncoder().fit_transform(X[:, i])
        f_sizes.append(int(X[:, i].max() + 1))
        Xs.append(X[:, i])

    print(f_sizes)
    fm = KerasFM()
    model, model_features = fm.build_model(f_sizes)
    earlystopper = EarlyStopping(patience=0, verbose=1)
    model.fit(Xs, y.ravel(), epochs=10,
              batch_size=2 ** 7,
              verbose=1,
              shuffle=True,
              sample_weight=None,
              callbacks=[earlystopper],
              )
    df_ = fm.get_factors_biases(Xs, range(len(Xs)), model_features)
    print(df_)
    print(df_.shape) # n_features * 3

    print(model.predict(Xs))


#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""
__title__ = 'DataAugment'
__author__ = 'JieYuan'
__mtime__ = '19-3-21'
"""

import pandas as pd
from sklearn.utils import shuffle


class DataAugment(object):

    def __init__(self, num_n, num_p, seed=0):
        """
        :param num_n: Negative sample enhancement multiple
        :param num_p: Positive sample enhancement multiple
        :param seed:
        """
        self.num_n = num_n
        self.num_p = num_p
        self.seed = seed

    def transform(self, X: pd.DataFrame, y):
        if not hasattr(y, 'tolist'):
            y = pd.Series(y)

        cols = X.columns
        X.columns = range(len(cols))
        Xn = self._augment(X[y == 0], self.num_n, self.seed)
        Xp = self._augment(X[y == 1], self.num_p, self.seed + 666666)

        X_ = pd.concat([X] + Xn + Xp, ignore_index=True)
        X_.columns = cols

        y_ = y.tolist() + [0] * (y == 0).sum() * self.num_n + [1] * y.sum() * self.num_p
        return X_, pd.Series(y_)

    def _augment(self, X, num, seed=0):
        return [X.apply(lambda x: shuffle(x.values, random_state=x.name + seed + i)) for i in range(num)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : DeepFM
# @Time         : 2020/5/22 10:03 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : https://img-blog.csdn.net/2018050820141665?watermark/2/text/aHR0cHM6Ly9ibG9nLmNzZG4ubmV0L3NvbmdiaW54dQ==/font/5a6L5L2T/fontsize/400/fill/I0JBQkFCMA==/dissolve/70
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

from inn.models.BaseModel import BaseModel

import tensorflow as tf
from inn.features.fc import *

from inn.layers.DNN import DNN
from inn.layers.interaction import FM
from inn.features.fc.utils import get_linear_output

from inn.layers.EmbeddingPooling import EmbeddingPooling


class DeepFM(BaseModel):

    def __init__(self, fcs: List[Column],
                 fm_layer=FM(),
                 dnn_layer=DNN(),
                 num_class=2,
                 model_name='DeepFM',
                 **kwargs):
        super().__init__(fcs=fcs,
                         num_class=num_class,
                         model_name=model_name,
                         **kwargs)
        self.fm_layer = fm_layer
        self.dnn_layer = dnn_layer

    def _build_model(self, **kwargs):
        """
        sigmoid(concatenate(fm, dnn))
        线性部分：多数值dense(1) + 每个cat embedding(1)
        二阶：
        FM部分：FM(embedding(k))

        """
        nc_outs = []
        cc_outs = []
        for fc, input in self.fc2input:
            if isinstance(fc, NumericColumn):
                _ = tf.keras.layers.Dense(1, activation=None)(input)  # activation
                nc_outs.append(_)

            elif isinstance(fc, (CategoricalColumn, SequenceCategoricalColumn)):
                _ = tf.keras.layers.Embedding(fc.vocabulary_size, 1)(input)
                _1 = tf.keras.layers.GlobalAveragePooling1D()(_)

                _2 = tf.keras.layers.Embedding(fc.vocabulary_size, fc.embedding_dim)(input)

                cc_outs.append((_1, _2))

        linear_out = cc_outs + [i for i, j in cc_outs]
        if linear_out:
            if len(linear_out) > 1:
                linear_out = tf.keras.layers.concatenate(nc_outs)
            else:
                linear_out = linear_out[0]

        if cc_outs:
            if len(cc_outs) > 1:
                _ = tf.keras.layers.concatenate([j for i, j in cc_outs], axis=1)
                fm_out = self.fm_layer(_)

                _ = tf.keras.layers.concatenate([tf.keras.layers.GlobalAveragePooling1D()(j) for i, j in cc_outs])

            else:
                fm_out = self.fm_layer(cc_outs[0])







        dnn_out = self.dnn(k_out)
        dnn_out = tf.keras.layers.Dense(1, use_bias=False, activation=None)(dnn_out)

        # 合并输出
        out = tf.keras.layers.concatenate([fm_out, dnn_out])  # 为啥是＋
        out = self.prediction_output_layer(out)

        model = tf.keras.models.Model(inputs=self.inputs, outputs=out)
        return model


if __name__ == '__main__':
    fcs = [NumericColumn('num'), CategoricalColumn('cat')]

    print(DeepFM(fcs).summary())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : smooth_utils
# @Time         : 2020/5/20 10:40 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import numpy as np


def walson_ctr(num_click, num_pv, z=1.96):
    """:arg
    威尔逊
    https://mp.weixin.qq.com/s/rLP1wsS0a71q5RA7NQcjdQ
    """
    p = num_click / num_pv
    if p > 0.9:
        return 0.0

    n = num_pv

    A = p + z ** 2 / (2 * n)
    B = np.sqrt(p * (1 - p) / n + z ** 2 / (4 * (n ** 2)))
    C = z * B

    D = 1 + z ** 2 / n

    ctr = (A - C) / D

    return ctr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : __init__.py
# @Time         : 2020/4/13 11:41 上午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : https://github.com/brennv/namedtupled

from .namedtupled import mapper, ignore, reducer
from .integrations import load_json, load_yaml, load_lists, load_env

map = mapper
reduce = reducer
json = load_json
yaml = load_yaml
zip = load_lists
env = load_env
